from qgis.core import QgsSettings, QgsProject, \
    QgsVectorLayer, QgsFeature, QgsGeometry, QgsCoordinateReferenceSystem, \
    QgsRasterLayer
import os.path


class SingleFeatureLayer(object):
    """
    Class for generating a layer containing a single feautre (essentially
    used for mask :
        - in cid the project display the current municiapity with
        a mask
        - in map_munip the current province hide the other municipality
    """

    def __init__(
        self, layer_name, get_feature, custom_prop_key,
        layer_tree_position=None, style_file=None):
        self.layer_name = layer_name
        self.get_feature = get_feature
        self.custom_prop_key = custom_prop_key
        self.layer_tree_position  = layer_tree_position
        self.style_file = style_file

    def get(self):
        for lid, l in QgsProject.instance().mapLayers().items():
            if l.providerType() == "memory" and l.customProperty(self.custom_prop_key, False):
                return l
        return None

    def clear(self):
        layer = self.get()
        if layer is not None:
            QgsProject.instance().removeMapLayers([layer.id()])

    def create_or_update(self, *args, **kwargs):
        feature = self.get_feature(*args, **kwargs)
        if not feature:
            self.clear()
            return

        layer = self.get()
        if layer is None:
            l = QgsVectorLayer("MultiPolygon?crs=EPSG:32735", self.layer_name, "memory")
            l.setCustomProperty(self.custom_prop_key, True)

            new_feature = QgsFeature()
            new_feature.setId(1)
            new_feature.setGeometry(QgsGeometry(feature.geometry()))
            l.dataProvider().addFeatures([new_feature])

            l = QgsProject.instance().layerStore().addMapLayer(l)
            if self.layer_tree_position is not None:
                QgsProject.instance().layerTreeRoot().insertLayer(
                    self.layer_tree_position, l)

            if self.style_file:
                l.loadNamedStyle(os.path.join(os.path.dirname(__file__), "layer_style", self.style_file))
            return l
        else:
            layer.dataProvider().changeGeometryValues({1: feature.geometry()})
            return layer

class PAGGFUtils(object):
    def __init__(self, paggf_layers):
        self.layers = paggf_layers  # instance de LoadedLayers
        self.plugin_dir = os.path.dirname(__file__)

        self.current_province_layer = SingleFeatureLayer(
            'Province courante', self.get_current_province,
            'current_province_layer')

        def get_munip_for_mask(mun_id=None):
            if not mun_id:
                mun_id = self.get_current_municipality_id()
                if mun_id is None:
                    return None

            f = self.layers.get_feature('ref', 'municipalities', mun_id)
            if f is None:
                return None
            return f

        self.mask_layer = SingleFeatureLayer(
            'Commune courante', get_munip_for_mask, 'is_mask', 0, 'mask.qml')

    def get_municipality(self, municipality_id):
        return self.layers.get_feature(
            "ref", "municipalities", municipality_id)

    def get_province_by_munip_id(self, municipality_id):
        municipality = self.get_municipality(municipality_id)
        if municipality is not None:
            return self.layers.get_feature(
                "ref", "provinces", municipality['id_province'])

    def get_current_municipality_id(self):
        s = QgsSettings()
        v = s.value("PAGGF/current_municipality", None)
        if v == "":
            v = None
        return v

    def get_current_municipality(self):
        municipality_id = self.get_current_municipality_id()
        if municipality_id:
            return self.get_municipality(municipality_id)

    def get_current_province(self):
        municipality = self.get_current_municipality()
        if municipality:
            return self.get_province_by_munip_id(municipality['id'])

    def get_encoded_idc_parcels_for(self, municipality):
        return self.layers.get_features_by_cond(
            'gis', 'encoded_idc_parcels', #TOOD idc_parcels
            (lambda f: f['id_sub_hill_avenue'][:4] == municipality['id'])
        )

def update_background_layer():
    "(Re)load the background raster layer"
    s = QgsSettings()
    ortho = s.value("PAGGF/ortho", "")
    ok = False
    if ortho:
        bg_lyr = QgsRasterLayer(ortho, "Orthophoto")
        bg_lyr.setCrs(QgsCoordinateReferenceSystem("EPSG:32735"))
        ok = bg_lyr.isValid()
    if not ok:
        bg_lyr = QgsRasterLayer("crs=EPSG:3857&format&type=xyz&url=http://b.tile.openstreetmap.org/{z}/{x}/{y}.png&zmax=18&zmin=0", "Fond OSM", "wms")

    # remove all raster layers
    for l in QgsProject.instance().mapLayers().values():
        if isinstance(l, QgsRasterLayer):
            QgsProject.instance().removeMapLayer(l)
    # add to the end of the layer stack
    bg_lyr = QgsProject.instance().layerStore().addMapLayer(bg_lyr)
    QgsProject.instance().layerTreeRoot().addLayer(bg_lyr)

def get_background_layer():
    for l in QgsProject.instance().mapLayers().values():
        if isinstance(l, QgsRasterLayer):
            return l
